@@ -1,4 +1,4 @@
-/* Copyright (C) 2005-2019 Massachusetts Institute of Technology
+/* Copyright (C) 2005-2021 Massachusetts Institute of Technology
 %
 %  This program is free software; you can redistribute it and/or modify
 %  it under the terms of the GNU General Public License as published by
@@ -44,7 +44,8 @@ typedef double realnum;
 
 #define MEEP_MIN_OUTPUT_TIME 4.0 // output no more often than this many seconds
 
-extern int verbosity; // if 0, suppress all non-error messages from Meep; 1 is default, 2 is debug output
+extern int
+    verbosity; // if 0, suppress all non-error messages from Meep; 1 is default, 2 is debug output
 
 const double pi = 3.141592653589793238462643383276;
 
@@ -61,6 +62,9 @@ class h5file;
 // Defined in monitor.cpp
 void matrix_invert(std::complex<double> (&Vinv)[9], std::complex<double> (&V)[9]);
 
+// Defined in dft.cpp
+std::vector<double> linspace(double freq_min, double freq_max, size_t Nfreq);
+
 double pml_quadratic_profile(double, void *);
 
 /* generic base class, only used by subclassing: represents susceptibility
@@ -370,8 +374,6 @@ protected:
   realnum *sigmat; // 5*T transition-specific sigma-diagonal factors
 };
 
-class grace;
-
 // h5file.cpp: HDF5 file I/O.  Most users, if they use this
 // class at all, will only use the constructor to open the file, and
 // will otherwise use the fields::output_hdf5 functions.
@@ -571,12 +573,13 @@ public:
   realnum *conductivity[NUM_FIELD_COMPONENTS][5];
   realnum *condinv[NUM_FIELD_COMPONENTS][5]; // cache of 1/(1+conduct*dt/2)
   bool condinv_stale;                        // true if condinv needs to be recomputed
-  double *sig[5], *kap[5], *siginv[5];       // conductivity array for uPML
-  int sigsize[5];                            // conductivity array size
+  double *sig[6], *kap[6], *siginv[6];       // conductivity array for uPML
+  int sigsize[6];                            // conductivity array size
   grid_volume gv; // integer grid_volume that could be bigger than non-overlapping v below
   volume v;
   susceptibility *chiP[NUM_FIELD_TYPES]; // only E_stuff and H_stuff are used
-  double cost;  // The cost of this chunk's grid_volume as computed by split_by_cost and fragment_stats
+  double
+      cost; // The cost of this chunk's grid_volume as computed by split_by_cost and fragment_stats
 
   int refcount; // reference count of objects using this structure_chunk
 
@@ -606,10 +609,12 @@ public:
   void remove_susceptibilities();
 
   // monitor.cpp
-  double get_chi1inv_at_pt(component, direction, int idx, double omega = 0) const;
-  double get_chi1inv(component, direction, const ivec &iloc, double omega = 0) const;
-  double get_inveps(component c, direction d, const ivec &iloc, double omega = 0) const {
-    return get_chi1inv(c, d, iloc, omega);
+  std::complex<double> get_chi1inv_at_pt(component, direction, int idx, double frequency = 0) const;
+  std::complex<double> get_chi1inv(component, direction, const ivec &iloc,
+                                   double frequency = 0) const;
+  std::complex<double> get_inveps(component c, direction d, const ivec &iloc,
+                                  double frequency = 0) const {
+    return get_chi1inv(c, d, iloc, frequency);
   }
   double max_eps() const;
 
@@ -771,18 +776,20 @@ public:
   void load_chunk_layout(const std::vector<grid_volume> &gvs, boundary_region &br);
 
   // monitor.cpp
-  double get_chi1inv(component, direction, const ivec &origloc, double omega = 0,
-                     bool parallel = true) const;
-  double get_chi1inv(component, direction, const vec &loc, double omega = 0,
-                     bool parallel = true) const;
-  double get_inveps(component c, direction d, const ivec &origloc, double omega = 0) const {
-    return get_chi1inv(c, d, origloc, omega);
+  std::complex<double> get_chi1inv(component, direction, const ivec &origloc, double frequency = 0,
+                                   bool parallel = true) const;
+  std::complex<double> get_chi1inv(component, direction, const vec &loc, double frequency = 0,
+                                   bool parallel = true) const;
+  std::complex<double> get_inveps(component c, direction d, const ivec &origloc,
+                                  double frequency = 0) const {
+    return get_chi1inv(c, d, origloc, frequency);
   }
-  double get_inveps(component c, direction d, const vec &loc, double omega = 0) const {
-    return get_chi1inv(c, d, loc, omega);
+  std::complex<double> get_inveps(component c, direction d, const vec &loc,
+                                  double frequency = 0) const {
+    return get_chi1inv(c, d, loc, frequency);
   }
-  double get_eps(const vec &loc, double omega = 0) const;
-  double get_mu(const vec &loc, double omega = 0) const;
+  std::complex<double> get_eps(const vec &loc, double frequency = 0) const;
+  std::complex<double> get_mu(const vec &loc, double frequency = 0) const;
   double max_eps() const;
   double estimated_cost(int process = my_rank());
 
@@ -800,6 +807,12 @@ private:
   void write_susceptibility_params(h5file *file, const char *dname, int EorH);
 };
 
+// defined in structure.cpp
+std::vector<grid_volume> choose_chunkdivision(grid_volume &gv,
+                                              volume &v,
+                                              int num_chunks,
+                                              const symmetry &s);
+
 class src_vol;
 class fields;
 class fields_chunk;
@@ -1007,8 +1020,7 @@ public:
   void operator-=(const dft_chunk &chunk);
 
   // the frequencies to loop_in_chunks
-  double omega_min, domega;
-  int Nomega;
+  std::vector<double> omega;
 
   component c; // component to DFT (possibly transformed by symmetry)
 
@@ -1082,6 +1094,12 @@ public:
   dft_flux(const component cE_, const component cH_, dft_chunk *E_, dft_chunk *H_, double fmin,
            double fmax, int Nf, const volume &where_, direction normal_direction_,
            bool use_symmetry_);
+  dft_flux(const component cE_, const component cH_, dft_chunk *E_, dft_chunk *H_,
+           const std::vector<double> freq_, const volume &where_, direction normal_direction_,
+           bool use_symmetry_);
+  dft_flux(const component cE_, const component cH_, dft_chunk *E_, dft_chunk *H_,
+           const double *freq_, size_t Nfreq, const volume &where_, direction normal_direction_,
+           bool use_symmetry_);
   dft_flux(const dft_flux &f);
 
   double *flux();
@@ -1101,8 +1119,7 @@ public:
 
   void remove();
 
-  double freq_min, dfreq;
-  int Nfreq;
+  std::vector<double> freq;
   dft_chunk *E, *H;
   component cE, cH;
   volume where;
@@ -1113,8 +1130,12 @@ public:
 // dft.cpp (normally created with fields::add_dft_energy)
 class dft_energy {
 public:
-  dft_energy(dft_chunk *E_, dft_chunk *H_, dft_chunk *D_, dft_chunk *B_, double fmin, double fmax,
-             int Nf, const volume &where_);
+  dft_energy(dft_chunk *E_, dft_chunk *H_, dft_chunk *D_, dft_chunk *B_, double freq_min,
+             double freq_max, int Nf, const volume &where_);
+  dft_energy(dft_chunk *E_, dft_chunk *H_, dft_chunk *D_, dft_chunk *B_,
+             const std::vector<double> freq_, const volume &where_);
+  dft_energy(dft_chunk *E_, dft_chunk *H_, dft_chunk *D_, dft_chunk *B_, const double *freq_,
+             size_t Nfreq, const volume &where_);
   dft_energy(const dft_energy &f);
 
   double *electric();
@@ -1138,8 +1159,7 @@ public:
 
   void remove();
 
-  double freq_min, dfreq;
-  int Nfreq;
+  std::vector<double> freq;
   dft_chunk *E, *H, *D, *B;
   volume where;
 };
@@ -1149,6 +1169,10 @@ class dft_force {
 public:
   dft_force(dft_chunk *offdiag1_, dft_chunk *offdiag2_, dft_chunk *diag_, double fmin, double fmax,
             int Nf, const volume &where_);
+  dft_force(dft_chunk *offdiag1_, dft_chunk *offdiag2_, dft_chunk *diag_,
+            const std::vector<double> freq_, const volume &where_);
+  dft_force(dft_chunk *offdiag1_, dft_chunk *offdiag2_, dft_chunk *diag_, const double *freq_,
+            size_t Nfreq, const volume &where_);
   dft_force(const dft_force &f);
 
   double *force();
@@ -1165,12 +1189,20 @@ public:
 
   void remove();
 
-  double freq_min, dfreq;
-  int Nfreq;
+  std::vector<double> freq;
   dft_chunk *offdiag1, *offdiag2, *diag;
   volume where;
 };
 
+
+struct sourcedata{
+  component near_fd_comp;
+  std::vector<ptrdiff_t> idx_arr;
+  int fc_idx;
+  std::vector<std::complex<double> > amp_arr;
+};
+
+
 // near2far.cpp (normally created with fields::add_dft_near2far)
 class dft_near2far {
 public:
@@ -1179,6 +1211,12 @@ public:
   dft_near2far(dft_chunk *F, double fmin, double fmax, int Nf, double eps, double mu,
                const volume &where_, const direction periodic_d_[2], const int periodic_n_[2],
                const double periodic_k_[2], const double period_[2]);
+  dft_near2far(dft_chunk *F, const std::vector<double> freq_, double eps, double mu,
+               const volume &where_, const direction periodic_d_[2], const int periodic_n_[2],
+               const double periodic_k_[2], const double period_[2]);
+  dft_near2far(dft_chunk *F, const double *freq_, size_t Nfreq, double eps, double mu,
+               const volume &where_, const direction periodic_d_[2], const int periodic_n_[2],
+               const double periodic_k_[2], const double period_[2]);
   dft_near2far(const dft_near2far &f);
 
   /* return an array (Ex,Ey,Ez,Hx,Hy,Hz) x Nfreq of the far fields at x */
@@ -1213,14 +1251,15 @@ public:
 
   void remove();
 
-  double freq_min, dfreq;
-  int Nfreq;
+  std::vector<double> freq;
   dft_chunk *F;
   double eps, mu;
   volume where;
   direction periodic_d[2];
   int periodic_n[2];
   double periodic_k[2], period[2];
+
+  std::vector<sourcedata> near_sourcedata(const vec &x_0, double* farpt_list, size_t nfar_pts, std::complex<double>* dJ);
 };
 
 /* Class to compute local-density-of-states spectra: the power spectrum
@@ -1232,6 +1271,8 @@ public:
 class dft_ldos {
 public:
   dft_ldos(double freq_min, double freq_max, int Nfreq);
+  dft_ldos(const std::vector<double> freq);
+  dft_ldos(const double *freq, size_t Nfreq);
   ~dft_ldos() {
     delete[] Fdft;
     delete[] Jdft;
@@ -1241,27 +1282,26 @@ public:
   double *ldos() const;            // returns array of Nomega values (after last timestep)
   std::complex<double> *F() const; // returns Fdft
   std::complex<double> *J() const; // returns Jdft
+  std::vector<double> freq;
 
 private:
   std::complex<realnum> *Fdft; // Nomega array of field * J*(x) DFT values
   std::complex<realnum> *Jdft; // Nomega array of J(t) DFT values
   double Jsum;                 // sum of |J| over all points
-public:
-  double omega_min, domega;
-  int Nomega;
 };
 
 // dft.cpp (normally created with fields::add_dft_fields)
 class dft_fields {
 public:
   dft_fields(dft_chunk *chunks, double freq_min, double freq_max, int Nfreq, const volume &where);
+  dft_fields(dft_chunk *chunks, const std::vector<double> freq_, const volume &where);
+  dft_fields(dft_chunk *chunks, const double *freq_, size_t Nfreq, const volume &where);
 
   void scale_dfts(std::complex<double> scale);
 
   void remove();
 
-  double freq_min, dfreq;
-  int Nfreq;
+  std::vector<double> freq;
   dft_chunk *chunks;
   volume where;
 };
@@ -1323,11 +1363,12 @@ public:
   structure_chunk *new_s;
   structure_chunk *s;
   const char *outdir;
+  int chunk_idx;
 
   fields_chunk(structure_chunk *, const char *outdir, double m, double beta,
-               bool zero_fields_near_cylorigin);
+               bool zero_fields_near_cylorigin, int chunkidx);
 
-  fields_chunk(const fields_chunk &);
+  fields_chunk(const fields_chunk &, int chunkidx);
   ~fields_chunk();
 
   void use_real_fields();
@@ -1344,7 +1385,8 @@ public:
   // monitor.cpp
   std::complex<double> get_field(component, const ivec &) const;
 
-  double get_chi1inv(component, direction, const ivec &iloc, double omega = 0) const;
+  std::complex<double> get_chi1inv(component, direction, const ivec &iloc,
+                                   double frequency = 0) const;
 
   void backup_component(component c);
   void average_with_backup(component c);
@@ -1460,6 +1502,39 @@ public:
   std::vector<std::complex<double> > values; // updated by update_values(idx)
 };
 
+class diffractedplanewave {
+
+public:
+  diffractedplanewave(int g[3], double axis[3], std::complex<double> s, std::complex<double> p);
+  int *get_g() { return g; };
+  double *get_axis() { return axis; }
+  std::complex<double> get_s() const { return s; };
+  std::complex<double> get_p() const { return p; };
+
+private:
+  int g[3];                // diffraction order
+  double axis[3];          // axis vector
+  std::complex<double> s;  // s polarization amplitude
+  std::complex<double> p;  // p polarization ampiltude
+};
+
+class gaussianbeam {
+
+public:
+  gaussianbeam(const vec &x0, const vec &kdir, double w0, double freq,
+               double eps, double mu, std::complex<double> E0[3]);
+  void get_fields(std::complex<double> *EH, const vec &x) const;
+  std::complex<double> get_E0(int n) const { return E0[n]; };
+
+private:
+  vec x0;           // beam center
+  vec kdir;         // beam propagation direction
+  double w0;        // beam waist radius
+  double freq;      // beam frequency
+  double eps, mu;   // permittivity/permeability of homogeneous medium
+  std::complex<double> E0[3]; // polarization vector
+};
+
 /***************************************************************/
 /* prototype for optional user-supplied function to provide an */
 /* initial estimate of the wavevector of mode #mode at         */
@@ -1514,7 +1589,7 @@ public:
   void reset();
 
   // time.cpp
-  double time_spent_on(time_sink);
+  std::vector<double> time_spent_on(time_sink);
   double mean_time_spent_on(time_sink);
   void print_times();
   // boundaries.cpp
@@ -1532,27 +1607,29 @@ public:
   // low-level function:
   void output_hdf5(h5file *file, const char *dataname, int num_fields, const component *components,
                    field_function fun, void *fun_data_, int reim, const volume &where,
-                   bool append_data = false, bool single_precision = false, double omega = 0);
+                   bool append_data = false, bool single_precision = false, double frequency = 0);
   // higher-level functions
   void output_hdf5(const char *dataname, // OUTPUT COMPLEX-VALUED FUNCTION
                    int num_fields, const component *components, field_function fun, void *fun_data_,
                    const volume &where, h5file *file = 0, bool append_data = false,
                    bool single_precision = false, const char *prefix = 0,
-                   bool real_part_only = false, double omega = 0);
+                   bool real_part_only = false, double frequency = 0);
   void output_hdf5(const char *dataname, // OUTPUT REAL-VALUED FUNCTION
                    int num_fields, const component *components, field_rfunction fun,
                    void *fun_data_, const volume &where, h5file *file = 0, bool append_data = false,
                    bool single_precision = false, const char *prefix = 0, double = 0);
   void output_hdf5(component c, // OUTPUT FIELD COMPONENT (or Dielectric)
                    const volume &where, h5file *file = 0, bool append_data = false,
-                   bool single_precision = false, const char *prefix = 0, double omega = 0);
+                   bool single_precision = false, const char *prefix = 0, double frequency = 0);
   void output_hdf5(derived_component c, // OUTPUT DERIVED FIELD COMPONENT
                    const volume &where, h5file *file = 0, bool append_data = false,
-                   bool single_precision = false, const char *prefix = 0, double omega = 0);
+                   bool single_precision = false, const char *prefix = 0, double frequency = 0);
   h5file *open_h5file(const char *name, h5file::access_mode mode = h5file::WRITE,
                       const char *prefix = NULL, bool timestamp = false);
   const char *h5file_name(const char *name, const char *prefix = NULL, bool timestamp = false);
 
+  void output_times(const char *fname);
+
   // array_slice.cpp methods
 
   // given a subvolume, compute the dimensions of the array slice
@@ -1563,24 +1640,15 @@ public:
   // in the x direction (corresponding to the two grid points
   // nearest x0, from which fields at x0 are interpolated).
   // if collapse_empty_dimensions==true, all such length-2
-  // array dimensions are collaped to length 1 by doing the
+  // array dimensions are collapsed to length 1 by doing the
   // interpolation before returning the array.
-  // currently, collapse_empty_dimensions is always false for the
-  // time-domain arrays returned by get_field_array and always
-  // true for the frequency-domain arrays returned by get_dft_array,
-  // so an alternative name for `collapse_empty_dimensions` would be
-  // `is_dft_array`.
   //
   // the `data` parameter is used internally in get_array_slice
   // and should be ignored by external callers.
   int get_array_slice_dimensions(const volume &where, size_t dims[3], direction dirs[3],
                                  bool collapse_empty_dimensions = false,
                                  bool snap_empty_dimensions = false, vec *min_max_loc = NULL,
-                                 void *data = 0);
-
-  int get_dft_array_dimensions(const volume &where, size_t dims[3], direction dirs[3]) {
-    return get_array_slice_dimensions(where, dims, dirs, true);
-  }
+                                 void *data = 0, component cgrid = Centered);
 
   // given a subvolume, return a column-major array containing
   // the given function of the field components in that subvolume
@@ -1590,21 +1658,27 @@ public:
   // must eventually be caller-deallocated via delete[].
   double *get_array_slice(const volume &where, std::vector<component> components,
                           field_rfunction rfun, void *fun_data, double *slice = 0,
-                          double omega = 0);
+                          double frequency = 0,
+                          bool snap = false);
 
   std::complex<double> *get_complex_array_slice(const volume &where,
                                                 std::vector<component> components,
                                                 field_function fun, void *fun_data,
-                                                std::complex<double> *slice = 0, double omega = 0);
+                                                std::complex<double> *slice = 0,
+                                                double frequency = 0,
+                                                bool snap = false);
 
   // alternative entry points for when you have no field
   // function, i.e. you want just a single component or
   // derived component.)
-  double *get_array_slice(const volume &where, component c, double *slice = 0, double omega = 0);
+  double *get_array_slice(const volume &where, component c, double *slice = 0,
+                          double frequency = 0, bool snap = false);
   double *get_array_slice(const volume &where, derived_component c, double *slice = 0,
-                          double omega = 0);
+                          double frequency = 0, bool snap = false);
   std::complex<double> *get_complex_array_slice(const volume &where, component c,
-                                                std::complex<double> *slice = 0, double omega = 0);
+                                                std::complex<double> *slice = 0,
+                                                double frequency = 0,
+                                                bool snap = false);
 
   // like get_array_slice, but for *sources* instead of fields
   std::complex<double> *get_source_slice(const volume &where, component source_slice_component,
@@ -1613,13 +1687,11 @@ public:
   // master routine for all above entry points
   void *do_get_array_slice(const volume &where, std::vector<component> components,
                            field_function fun, field_rfunction rfun, void *fun_data, void *vslice,
-                           double omega = 0);
+                           double frequency = 0, bool snap = false);
 
-  /* fetch and return coordinates and integration weights of grid points covered by an array slice,
-   */
+  /* fetch and return coordinates and integration weights of grid points covered by an array slice, */
   /* packed into a vector with format [NX, xtics[:], NY, ytics[:], NZ, ztics[:], weights[:] ] */
-  std::vector<double> get_array_metadata(const volume &where, bool collapse_empty_dimensions = true,
-                                         bool snap_empty_dimensions = false);
+  std::vector<double> get_array_metadata(const volume &where);
 
   // step.cpp methods:
   double last_step_output_wall_time;
@@ -1633,8 +1705,10 @@ public:
   inline double time() const { return t * dt; };
 
   // cw_fields.cpp:
-  bool solve_cw(double tol, int maxiters, std::complex<double> frequency, int L = 2);
-  bool solve_cw(double tol = 1e-8, int maxiters = 10000, int L = 2);
+  bool solve_cw(double tol, int maxiters, std::complex<double> frequency, int L = 2,
+                std::complex<double> *eigfreq = NULL, double eigtol = 1e-8, int eigiters = 20);
+  bool solve_cw(double tol = 1e-8, int maxiters = 10000, int L = 2,
+                std::complex<double> *eigfreq = NULL, double eigtol = 1e-8, int eigiters = 20);
 
   // sources.cpp:
   double last_source_time();
@@ -1652,7 +1726,9 @@ public:
                          std::complex<double> A(const vec &), std::complex<double> amp = 1.0);
   void add_volume_source(component c, const src_time &src, const volume &,
                          std::complex<double> amp = 1.0);
+  void add_volume_source(const src_time &src, const volume &, gaussianbeam beam);
   void require_component(component c);
+  void add_srcdata(struct sourcedata cur_data, src_time *src, size_t n, std::complex<double>* amp_arr);
 
   // mpb.cpp
 
@@ -1660,27 +1736,30 @@ public:
   // that can be passed to eigenmode_amplitude() to get
   // values of field components at arbitrary points in space.
   // call destroy_eigenmode_data() to deallocate it when finished.
-  void *get_eigenmode(double omega_src, direction d, const volume where, const volume eig_vol,
+  void *get_eigenmode(double frequency, direction d, const volume where, const volume eig_vol,
                       int band_num, const vec &kpoint, bool match_frequency, int parity,
-                      double resolution, double eigensolver_tol,
-                      double *kdom = 0, void **user_mdata = 0);
+                      double resolution, double eigensolver_tol, double *kdom = 0,
+                      void **user_mdata = 0, diffractedplanewave *dp = 0);
 
   void add_eigenmode_source(component c, const src_time &src, direction d, const volume &where,
                             const volume &eig_vol, int band_num, const vec &kpoint,
                             bool match_frequency, int parity, double eig_resolution,
                             double eigensolver_tol, std::complex<double> amp,
-                            std::complex<double> A(const vec &) = 0);
+                            std::complex<double> A(const vec &) = 0,
+                            diffractedplanewave *dp = 0);
 
   void get_eigenmode_coefficients(dft_flux flux, const volume &eig_vol, int *bands, int num_bands,
                                   int parity, double eig_resolution, double eigensolver_tol,
                                   std::complex<double> *coeffs, double *vgrp,
                                   kpoint_func user_kpoint_func, void *user_kpoint_data,
-                                  vec *kpoints, vec *kdom, direction d);
+                                  vec *kpoints, vec *kdom, double *cscale, direction d,
+                                  diffractedplanewave *dp = 0);
   void get_eigenmode_coefficients(dft_flux flux, const volume &eig_vol, int *bands, int num_bands,
                                   int parity, double eig_resolution, double eigensolver_tol,
                                   std::complex<double> *coeffs, double *vgrp,
                                   kpoint_func user_kpoint_func = 0, void *user_kpoint_data = 0,
-                                  vec *kpoints = 0, vec *kdom = 0);
+                                  vec *kpoints = 0, vec *kdom = 0, double *cscale = 0,
+                                  diffractedplanewave *dp = 0);
 
   // initialize.cpp:
   void initialize_field(component, std::complex<double> f(const vec &));
@@ -1720,28 +1799,92 @@ public:
 
   // dft.cpp
   dft_chunk *add_dft(component c, const volume &where, double freq_min, double freq_max, int Nfreq,
+                     bool include_dV_and_interp_weights = true,
+                     std::complex<double> stored_weight = 1.0, dft_chunk *chunk_next = 0,
+                     bool sqrt_dV_and_interp_weights = false,
+                     std::complex<double> extra_weight = 1.0, bool use_centered_grid = true,
+                     int vc = 0) {
+    return add_dft(c, where, linspace(freq_min, freq_max, Nfreq), include_dV_and_interp_weights,
+                   stored_weight, chunk_next, sqrt_dV_and_interp_weights, extra_weight,
+                   use_centered_grid, vc);
+  }
+  dft_chunk *add_dft(component c, const volume &where, const double *freq, size_t Nfreq,
                      bool include_dV_and_interp_weights = true,
                      std::complex<double> stored_weight = 1.0, dft_chunk *chunk_next = 0,
                      bool sqrt_dV_and_interp_weights = false,
                      std::complex<double> extra_weight = 1.0, bool use_centered_grid = true,
                      int vc = 0);
-  dft_chunk *add_dft_pt(component c, const vec &where, double freq_min, double freq_max, int Nfreq);
+  dft_chunk *add_dft(component c, const volume &where, const std::vector<double> freq,
+                     bool include_dV_and_interp_weights = true,
+                     std::complex<double> stored_weight = 1.0, dft_chunk *chunk_next = 0,
+                     bool sqrt_dV_and_interp_weights = false,
+                     std::complex<double> extra_weight = 1.0, bool use_centered_grid = true,
+                     int vc = 0) {
+    return add_dft(c, where, freq.data(), freq.size(), include_dV_and_interp_weights, stored_weight,
+                   chunk_next, sqrt_dV_and_interp_weights, extra_weight, use_centered_grid, vc);
+  }
+  dft_chunk *add_dft_pt(component c, const vec &where, double freq_min, double freq_max,
+                        int Nfreq) {
+    return add_dft(c, where, linspace(freq_min, freq_max, Nfreq), false);
+  }
+  dft_chunk *add_dft_pt(component c, const vec &where, const std::vector<double> freq) {
+    return add_dft(c, where, freq, false);
+  }
   dft_chunk *add_dft(const volume_list *where, double freq_min, double freq_max, int Nfreq,
+                     bool include_dV = true) {
+    return add_dft(where, linspace(freq_min, freq_max, Nfreq), include_dV);
+  }
+  dft_chunk *add_dft(const volume_list *where, const std::vector<double> freq,
                      bool include_dV = true);
   void update_dfts();
+  dft_flux add_dft_flux(const volume_list *where, const double *freq, size_t Nfreq,
+                        bool use_symmetry = true, bool centered_grid = true);
+  dft_flux add_dft_flux(const volume_list *where, const std::vector<double> freq,
+                        bool use_symmetry = true, bool centered_grid = true) {
+    return add_dft_flux(where, freq.data(), freq.size(), use_symmetry, centered_grid);
+  }
   dft_flux add_dft_flux(const volume_list *where, double freq_min, double freq_max, int Nfreq,
-                        bool use_symmetry = true);
+                        bool use_symmetry = true, bool centered_grid = true) {
+    return add_dft_flux(where, linspace(freq_min, freq_max, Nfreq), use_symmetry, centered_grid);
+  }
   dft_flux add_dft_flux(direction d, const volume &where, double freq_min, double freq_max,
-                        int Nfreq, bool use_symmetry = true);
+                        int Nfreq, bool use_symmetry = true, bool centered_grid = true) {
+    return add_dft_flux(d, where, linspace(freq_min, freq_max, Nfreq), use_symmetry, centered_grid);
+  }
+  dft_flux add_dft_flux(direction d, const volume &where, const std::vector<double> freq,
+                        bool use_symmetry = true, bool centered_grid = true) {
+    return add_dft_flux(d, where, freq.data(), freq.size(), use_symmetry, centered_grid);
+  }
+  dft_flux add_dft_flux(direction d, const volume &where, const double *freq, size_t Nfreq,
+                        bool use_symmetry = true, bool centered_grid = true);
   dft_flux add_dft_flux_box(const volume &where, double freq_min, double freq_max, int Nfreq);
+  dft_flux add_dft_flux_box(const volume &where, const std::vector<double> freq);
   dft_flux add_dft_flux_plane(const volume &where, double freq_min, double freq_max, int Nfreq);
+  dft_flux add_dft_flux_plane(const volume &where, const std::vector<double> freq);
 
   // a "mode monitor" is just a dft_flux with symmetry reduction turned off.
   dft_flux add_mode_monitor(direction d, const volume &where, double freq_min, double freq_max,
-                            int Nfreq);
+                            int Nfreq, bool centered_grid = true) {
+    return add_mode_monitor(d, where, linspace(freq_min, freq_max, Nfreq), centered_grid);
+  }
+  dft_flux add_mode_monitor(direction d, const volume &where, const std::vector<double> freq, bool centered_grid = true) {
+    return add_mode_monitor(d, where, freq.data(), freq.size(), centered_grid);
+  }
+  dft_flux add_mode_monitor(direction d, const volume &where, const double *freq, size_t Nfreq, bool centered_grid = true);
 
   dft_fields add_dft_fields(component *components, int num_components, const volume where,
-                            double freq_min, double freq_max, int Nfreq);
+                            double freq_min, double freq_max, int Nfreq,
+                            bool use_centered_grid = true) {
+    return add_dft_fields(components, num_components, where, linspace(freq_min, freq_max, Nfreq),
+                          use_centered_grid);
+  }
+  dft_fields add_dft_fields(component *components, int num_components, const volume where,
+                            const std::vector<double> freq, bool use_centered_grid = true) {
+    return add_dft_fields(components, num_components, where, freq.data(), freq.size(),
+                          use_centered_grid);
+  }
+  dft_fields add_dft_fields(component *components, int num_components, const volume where,
+                            const double *freq, size_t Nfreq, bool use_centered_grid = true);
 
   /********************************************************/
   /* process_dft_component is an intermediate-level       */
@@ -1767,7 +1910,6 @@ public:
   void output_dft(dft_force force, const char *HDF5FileName);
   void output_dft(dft_near2far n2f, const char *HDF5FileName);
   void output_dft(dft_fields fdft, const char *HDF5FileName);
-  void output_mode_fields(void *mode_data, dft_flux flux, const char *HDF5FileName);
 
   // get array of DFT field values
   std::complex<double> *get_dft_array(dft_flux flux, component c, int num_freq, int *rank,
@@ -1787,22 +1929,43 @@ public:
   void get_mode_mode_overlap(void *mode1_data, void *mode2_data, dft_flux flux,
                              std::complex<double> overlaps[2]);
 
-  dft_energy add_dft_energy(const volume_list *where, double freq_min, double freq_max, int Nfreq);
+  dft_energy add_dft_energy(const volume_list *where, double freq_min, double freq_max, int Nfreq) {
+    return add_dft_energy(where, linspace(freq_min, freq_max, Nfreq));
+  }
+  dft_energy add_dft_energy(const volume_list *where, const std::vector<double> freq) {
+    return add_dft_energy(where, freq.data(), freq.size());
+  }
+  dft_energy add_dft_energy(const volume_list *where, const double *freq, size_t Nfreq);
 
   // stress.cpp
-  dft_force add_dft_force(const volume_list *where, double freq_min, double freq_max, int Nfreq);
+  dft_force add_dft_force(const volume_list *where, double freq_min, double freq_max, int Nfreq) {
+    return add_dft_force(where, linspace(freq_min, freq_max, Nfreq));
+  }
+  dft_force add_dft_force(const volume_list *where, const std::vector<double> freq) {
+    return add_dft_force(where, freq.data(), freq.size());
+  }
+  dft_force add_dft_force(const volume_list *where, const double *freq, size_t Nfreq);
 
   // near2far.cpp
   dft_near2far add_dft_near2far(const volume_list *where, double freq_min, double freq_max,
-                                int Nfreq, int Nperiods = 1);
-  // monitor.cpp
-  double get_chi1inv(component, direction, const vec &loc, double omega = 0,
-                     bool parallel = true) const;
-  double get_inveps(component c, direction d, const vec &loc, double omega = 0) const {
-    return get_chi1inv(c, d, loc, omega);
+                                int Nfreq, int Nperiods = 1) {
+    return add_dft_near2far(where, linspace(freq_min, freq_max, Nfreq), Nperiods);
   }
-  double get_eps(const vec &loc, double omega = 0) const;
-  double get_mu(const vec &loc, double omega = 0) const;
+  dft_near2far add_dft_near2far(const volume_list *where, const std::vector<double> freq,
+                                int Nperiods = 1) {
+    return add_dft_near2far(where, freq.data(), freq.size(), Nperiods);
+  }
+  dft_near2far add_dft_near2far(const volume_list *where, const double *freq, size_t Nfreq,
+                                int Nperiods = 1);
+  // monitor.cpp
+  std::complex<double> get_chi1inv(component, direction, const vec &loc, double frequency = 0,
+                                   bool parallel = true) const;
+  std::complex<double> get_inveps(component c, direction d, const vec &loc,
+                                  double frequency = 0) const {
+    return get_chi1inv(c, d, loc, frequency);
+  }
+  std::complex<double> get_eps(const vec &loc, double frequency = 0) const;
+  std::complex<double> get_mu(const vec &loc, double frequency = 0) const;
   void get_point(monitor_point *p, const vec &) const;
   monitor_point *get_new_point(const vec &, monitor_point *p = NULL) const;
 
@@ -1878,12 +2041,16 @@ private:
   void step_source(field_type ft, bool including_integrated = false);
   void update_pols(field_type ft);
   void calc_sources(double tim);
+  // sources.cpp
+  void add_volume_source_check(component c, const src_time &src, const volume &where,
+                               std::complex<double> A(const vec &), std::complex<double> amp,
+                               component c0, direction d, int has_tm, int has_te);
 
 public:
   // monitor.cpp
   std::complex<double> get_field(component c, const ivec &iloc, bool parallel = true) const;
-  double get_chi1inv(component, direction, const ivec &iloc, double omega = 0,
-                     bool parallel = true) const;
+  std::complex<double> get_chi1inv(component, direction, const ivec &iloc, double frequency = 0,
+                                   bool parallel = true) const;
   // boundaries.cpp
   bool locate_component_point(component *, ivec *, std::complex<double> *) const;
   // time.cpp
@@ -1929,16 +2096,11 @@ private:
 // directory, unless the source file hasn't changed.
 
 const char *make_output_directory(const char *exename, const char *jobname = NULL);
+char *make_output_directory(); // make temporary directory
 void trash_output_directory(const char *dirname);
+void delete_directory(const char *path);
 FILE *create_output_file(const char *dirname, const char *fname);
 
-// The following allows you to hit ctrl-C to tell your calculation to stop
-// and clean up.
-void deal_with_ctrl_c(int stop_now = 2);
-// When a ctrl_c is called, the following variable (which starts with a
-// zero value) is incremented.
-extern int interrupt;
-
 int do_harminv(std::complex<double> *data, int n, double dt, double fmin, double fmax, int maxbands,
                std::complex<double> *amps, double *freq_re, double *freq_im, double *errors = NULL,
                double spectral_density = 1.1, double Q_thresh = 50, double rel_err_thresh = 1e20,
@@ -1958,6 +2120,7 @@ void broadcast(int from, realnum *data, int size);
 
 // random number generation: random.cpp
 void set_random_seed(unsigned long seed);
+void restore_random_seed();
 double uniform_random(double a, double b);          // uniform random in [a,b]
 double gaussian_random(double mean, double stddev); // normal random with given mean and stddev
 int random_int(int a, int b);                       // uniform random in [a,b)
@@ -1983,12 +2146,6 @@ vec get_k(void *vedata);
 realnum linear_interpolate(realnum rx, realnum ry, realnum rz, realnum *data, int nx, int ny,
                            int nz, int stride);
 
-// utility routine for modular arithmetic that always returns a nonnegative integer
-inline int pmod(int n, int modulus) {
-  n = n % modulus;
-  if (n < 0) n += modulus;
-  return n;
-}
 } /* namespace meep */
 
 #endif /* MEEP_H */
